"""ZenHub release reports methods."""
import datetime
from typing import Iterable, List, Optional, Union

from ..models import ReleaseReport
from ..types import Base64String, ReportState
from ..utils import check_dates, date_to_string
from .base import BaseMixin


class ReleaseReportsMixin(BaseMixin):
    def create_release_report(
        self,
        repo_id: int,
        title: str,
        start_date: datetime.datetime,
        desired_end_date: datetime.datetime,
        description: str = "",
        repositories: Iterable[int] = (),
    ) -> Union[ReleaseReport, dict]:
        """
        Create a Release Report.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        title : str
            Title of the Release Report.
        start_date : datetime.datetime
            Start date of the Release Report.
        desired_end_date : datetime.datetime
            End date of the Release Report.
        description : str, optional
            Start date of the Release Report. The default is ''.
        repositories : Iterable of int, optional
            List of repository IDs to include in the Release Report.
            The default is ().

        Returns
        -------
        ReleaseReport or dict
            The created Release Report. See example response below.

        .. code-block:: python
            {
                "release_id": "59dff4f508399a35a276a1ea",
                "title": "Great title",
                "description": "Amazing description",
                "start_date": "2007-01-01T00:00:00.000Z",
                "desired_end_date": "2007-01-01T00:00:00.000Z",
                "created_at": "2017-10-12T23:04:21.795Z",
                "closed_at": None,
                "state": "open",
                "repositories": [103707262]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#create-a-release-report
        """
        self._repo_id = repo_id
        check_dates(start_date, desired_end_date)
        # POST /p1/repositories/:repo_id/reports/release
        url = f"/p1/repositories/{repo_id}/reports/release"
        body = {
            "title": title,
            "start_date": date_to_string(start_date),
            "desired_end_date": date_to_string(desired_end_date),
        }
        if description:
            body["description"] = description

        if repositories:
            body["repositories"] = list(repositories)  # type: ignore [assignment]

        data = self._post(url, body)
        model = ReleaseReport.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def get_release_report(
        self, release_id: Base64String
    ) -> Union[ReleaseReport, dict]:
        """
        Get a Release Report.

        Parameters
        ----------
        release_id : Base64String
            The unique string identifier of the Release Report.

        Returns
        -------
        ReleaseReport or dict
            The requested Release Report. See example response below.

        .. code-block:: python
            {
                "release_id": "59d3cd520a430a6344fd3bdb",
                "title": "Test release",
                "description": "",
                "start_date": "2017-10-01T19:00:00.000Z",
                "desired_end_date": "2017-10-03T19:00:00.000Z",
                "created_at": "2017-10-03T17:48:02.701Z",
                "closed_at": None,
                "state": "open",
                "repositories": [105683718]
            }
        Note
        ----
        https://github.com/ZenHubIO/API#get-a-release-report
        """
        # GET /p1/reports/release/:release_id
        url = f"/p1/reports/release/{release_id}"
        data = self._get(url)
        model = ReleaseReport.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def get_release_reports(
        self, repo_id: int
    ) -> Union[List[ReleaseReport], List[dict]]:
        """
        Get Release Reports for a Repository.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        List of ReleaseReport or List of dict. See example response below.

        .. code-block:: python
            [
                {
                    "release_id": "59cbf2fde010f7a5207406e8",
                    "title": "Great title for release 1",
                    "description": "Great description for release",
                    "start_date": "2000-10-10T00:00:00.000Z",
                    "desired_end_date": "2010-10-10T00:00:00.000Z",
                    "created_at": "2017-09-27T18:50:37.418Z",
                    "closed_at": None,
                    "state": "open"
                },
                {
                    "release_id": "59cbf2fde010f7a5207406e8",
                    "title": "Great title for release 2",
                    "description": "Great description for release",
                    "start_date": "2000-10-10T00:00:00.000Z",
                    "desired_end_date": "2010-10-10T00:00:00.000Z",
                    "created_at": "2017-09-27T18:50:37.418Z",
                    "closed_at": None,
                    "state": "open"
                }
            ]

        Note
        ----
        https://github.com/ZenHubIO/API#get-release-reports-for-a-repository
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/reports/releases
        url = f"/p1/repositories/{repo_id}/reports/releases"
        if self._output_models:
            return [ReleaseReport.parse_obj(item) for item in self._get(url)]
        else:
            return [
                ReleaseReport.parse_obj(item).dict(include=item.keys())
                for item in self._get(url)
            ]

    def edit_release_report(
        self,
        release_id: Base64String,
        title: str,
        start_date: datetime.datetime,
        desired_end_date: datetime.datetime,
        description: str = '',
        state: Optional[ReportState] = None,
    ) -> Union[ReleaseReport, dict]:
        """
        Edit a Release Report.

        Parameters
        ----------
        release_id : Base64String
            The unique string identifier of the Release Report.
        title : str
            Title of the Release Report.
        start_date : datetime.datetime
            Start date of the Release Report.
        desired_end_date : datetime.datetime
            End date of the Release Report.
        description : str, optional
            Start date of the Release Report. The default is ''.
        state : ReportState, optional
            The state the Release Report. The default is None.

        Returns
        -------
        ReleaseReport or dict
            The created Release Report. See example response below.

        .. code-block:: python
            {
                "release_id": "59d3d6438b3f16667f9e7174",
                "title": "Amazing title",
                "description": "Amazing description",
                "start_date": "2007-01-01T00:00:00.000Z",
                "desired_end_date": "2007-01-01T00:00:00.000Z",
                "created_at": "2017-10-03T18:26:11.700Z",
                "closed_at": "2017-10-03T18:26:11.700Z",
                "state": "closed",
                "repositories": [105683567, 105683718]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#edit-a-release-report
        """
        check_dates(start_date, desired_end_date)
        # PATCH /p1/reports/release/:release_id
        url = f"/p1/reports/release/{release_id}"
        body = {
            "title": title,
            "description": description,
            "start_date": date_to_string(start_date),
            "desired_end_date": date_to_string(desired_end_date),
        }
        if state is not None:
            if state in ["open", "closed"]:
                body["state"] = state
            else:
                raise ValueError("`state` must be 'open' or 'closed'")

        data = self._patch(url, body)
        model = ReleaseReport.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def add_repo_to_release_report(
        self, release_id: Base64String, repo_id: int
    ) -> bool:
        """
        Add a Repository to a Release Report.

        Parameters
        ----------
        release_id : Base64String
            The unique string identifier of the Release Report.
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        bool
            ``True`` if successful.

        Note
        ----
        https://github.com/ZenHubIO/API#add-a-repository-to-a-release-report
        """
        self._repo_id = repo_id
        # POST /p1/reports/release/:release_id/repository/:repo_id
        url = f"/p1/reports/release/{release_id}/repository/{repo_id}"
        return True if self._post(url) == {} else False

    def remove_repo_from_release_report(
        self, release_id: Base64String, repo_id: int
    ) -> bool:
        """
        Remove a Repository from a Release Report.

        Parameters
        ----------
        release_id : Base64String
            The unique string identifier of the Release Report.
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        bool
            ``True`` if successful.

        Note
        ----
        A release must always have at least one repository.
        https://github.com/ZenHubIO/API#remove-a-repository-from-a-release-report
        """
        self._repo_id = repo_id
        # DELETE /p1/reports/release/:release_id/repository/:repo_id
        url = f"/p1/reports/release/{release_id}/repository/{repo_id}"
        return True if self._delete(url) == {} else False



from typing import Optional

import requests

from ..types import URLString
from ..utils import parse_response_contents


class BaseMixin:
    _session: requests.Session
    _base_url: URLString
    _repo_id: Optional[int]
    _output_models: bool

    def _make_url(self, url: URLString) -> URLString:
        """Create full api url."""
        return f"{self._base_url}{url}"

    def _get(self, url: URLString) -> dict:
        """Send GET request with given url."""
        response = self._session.get(url=self._make_url(url))
        return parse_response_contents(response)

    def _post(self, url: URLString, body: dict = {}) -> dict:
        """Send POST request with given url and data."""
        response = self._session.post(url=self._make_url(url), json=body)
        return parse_response_contents(response)

    def _put(self, url: URLString, body: dict) -> dict:
        """Send PUT request with given url and data."""
        response = self._session.put(url=self._make_url(url), json=body)
        return parse_response_contents(response)

    def _delete(self, url: URLString, body: dict = {}) -> dict:
        """Send DELETE request with given url and data."""
        response = self._session.delete(url=self._make_url(url), json=body)
        return parse_response_contents(response)

    def _patch(self, url: URLString, body: dict) -> dict:
        """Send PATCH request with given url and data."""
        response = self._session.patch(url=self._make_url(url), json=body)
        return parse_response_contents(response)



"""ZenHub API utilities."""
import datetime

import requests

from .exceptions import (
    APILimitError,
    InvalidTokenError,
    NotFoundError,
    ZenhubError,
)
from .types import ISO8601DateString


def date_to_string(date: datetime.datetime) -> ISO8601DateString:
    """Convert a datetime object to a ISO8601 date string."""
    return date.isoformat(timespec='milliseconds') + "Z"


def string_to_date(date_string: ISO8601DateString) -> datetime.datetime:
    """Convert a a ISO8601 date string to a datetime object."""
    if date_string.endswith("Z"):
        date_string = date_string[:-1]

    return datetime.datetime.fromisoformat(date_string)


def check_dates(
    start_date: datetime.datetime, desired_end_date: datetime.datetime
) -> bool:
    """Check ``desired_end_date`` comes after ``start_date``."""
    if start_date > desired_end_date:
        raise ValueError("Start date must be before end date.")
    return True


def parse_response_contents(response: requests.Response) -> dict:
    """Parse response and convert to json if possible."""
    status_code = response.status_code
    try:
        contents = response.json()
    except Exception:
        contents = {}

    if status_code in [200, 204]:
        pass
    elif status_code == 401:
        raise InvalidTokenError("Invalid token!")
    elif status_code == 403:
        raise APILimitError(
            "Reached request limit to the API. See API Limits."
        )
    elif status_code == 404:
        raise NotFoundError("Not found!")
    else:
        message = contents.get("message", "Unknown error!")
        raise ZenhubError(message)

    return contents



"""ZenHub workspace methods."""
from typing import List, Union

from ..models import Board, Workspace
from ..types import Base64String
from .base import BaseMixin


class WorkspacesMixin(BaseMixin):
    def get_workspaces(
        self, repo_id: int
    ) -> Union[List[dict], List[Workspace]]:
        """
        Gets all Workspaces containing ``repo_id``.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        List of dict.
            Zenhub list of workspaces. See example response below.

        .. code-block:: python
            [
                {
                    "name": "Design and UX",
                    "description": None,
                    "id": "5d0a7a9741fd098f6b7f58ac",
                    "repositories": [12345678, 912345]
                },
                {
                    "name": "Roadmap",
                    "description": "Feature planning and enhancements",
                    "id": "5d0a7cea41fd098f6b7f58b8",
                    "repositories": [12345678]
                }
            ]

        Note
        ----
        https://github.com/ZenHubIO/API#get-zenhub-workspaces-for-a-repository
        """
        self._repo_id = repo_id
        # GET /p2/repositories/:repo_id/workspaces
        url = f"/p2/repositories/{repo_id}/workspaces"
        data = self._get(url)
        if self._output_models:
            return [Workspace.parse_obj(workspace) for workspace in data]
        else:
            return [
                Workspace.parse_obj(workspace).dict(include=workspace.keys())
                for workspace in data
            ]

    def get_repository_board(
        self, workspace_id: Base64String, repo_id: int
    ) -> Union[dict, Board]:
        """
        Get ZenHub Board data for a repository (``repo_id``) within the
        Workspace (``workspace_id``).

        Parameters
        ----------
        workspace_id : Base64String
            Workspace unique string identifier.
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        Dictionary or model with board information.
            Zenhub workspace board listing pipelines and issues. See example
            response below.

        .. code-block:: python
            {
                "pipelines": [
                    {
                        "id": "595d430add03f01d32460080",
                        "name": "New Issues",
                        "issues": [
                            {
                                "issue_number": 279,
                                "estimate": {"value": 40},
                                "position": 0,
                                "is_epic": true
                            },
                            {
                                "issue_number": 142,
                                "is_epic": False
                            }
                        ]
                    },
                    {
                        "id": "595d430add03f01d32460081",
                        "name": "Backlog",
                        "issues": [
                            {
                                "issue_number": 303,
                                "estimate": {"value": 40},
                                "position": 3,
                                "is_epic": False
                            }
                        ]
                        },
                    {
                        "id": "595d430add03f01d32460082",
                        "name": "To Do",
                        "issues": [
                            {
                                "issue_number": 380,
                                "estimate": {"value": 1},
                                "position": 0,
                                "is_epic": True
                            },
                            {
                                "issue_number": 284,
                                "position": 2,
                                "is_epic": False
                            },
                            {
                                "issue_number": 329,
                                "estimate": {"value": 8},
                                "position": 7,
                                "is_epic": False
                            }
                        ]
                    }
                ]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#get-a-zenhub-board-for-a-repository
        """
        self._repo_id = repo_id
        # GET /p2/workspaces/:workspace_id/repositories/:repo_id/board
        url = f"/p2/workspaces/{workspace_id}/repositories/{repo_id}/board"
        data = self._get(url)
        model = Board.parse_obj(data)
        return (
            model
            if self._output_models
            else model.dict(include=data.keys(), exclude_none=True)
        )

    def get_oldest_repository_board(self, repo_id: int) -> Union[dict, Board]:
        """
        Get the oldest ZenHub board for a repository.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        Dictionary or model with board information.
            Zenhub workspace board listing pipelines and issues. See example
            response below.

        .. code-block:: python
            {
                "pipelines": [
                    {
                        "id": "595d430add03f01d32460080",
                        "name": "New Issues",
                        "issues": [
                            {
                                "issue_number": 279,
                                "estimate": {"value": 40},
                                "position": 0,
                                "is_epic": true
                            },
                            {
                                "issue_number": 142,
                                "is_epic": False
                            }
                        ]
                    },
                    {
                        "id": "595d430add03f01d32460081",
                        "name": "Backlog",
                        "issues": [
                            {
                                "issue_number": 303,
                                "estimate": {"value": 40},
                                "position": 3,
                                "is_epic": False
                            }
                        ]
                        },
                    {
                        "id": "595d430add03f01d32460082",
                        "name": "To Do",
                        "issues": [
                            {
                                "issue_number": 380,
                                "estimate": {"value": 1},
                                "position": 0,
                                "is_epic": True
                            },
                            {
                                "issue_number": 284,
                                "position": 2,
                                "is_epic": False
                            },
                            {
                                "issue_number": 329,
                                "estimate": {"value": 8},
                                "position": 7,
                                "is_epic": False
                            }
                        ]
                    }
                ]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#get-the-oldest-zenhub-board-for-a-repository
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/board
        url = f"/p1/repositories/{repo_id}/board"
        data = self._get(url)
        model = Board.parse_obj(data)
        return (
            model
            if self._output_models
            else model.dict(include=data.keys(), exclude_none=True)
        )



"""ZenHub API pydantic models."""
import datetime
from typing import List, Optional, Union

from pydantic import BaseModel

from .types import (
    Base64String,
    ISO8601DateString,
    IssueEventType,
    IssuePosition,
    ReportState,
    Seconds,
    URLString,
)


class IssueEstimate(BaseModel):
    value: int


class PipelineName(BaseModel):
    name: str


class Event(BaseModel):
    user_id: int
    type: IssueEventType
    created_at: datetime.datetime


class EstimateIssueEvent(Event):
    from_estimate: Optional[IssueEstimate]
    to_estimate: Optional[IssueEstimate]


class TransferIssueEvent(Event):
    from_pipeline: Optional[PipelineName]
    to_pipeline: Optional[PipelineName]
    workspace_id: Optional[Base64String]


class PlusOnes(BaseModel):
    created_at: ISO8601DateString


class Pipeline(BaseModel):
    name: str
    pipeline_id: Base64String
    workspace_id: Base64String


class IssueData(BaseModel):
    estimate: Optional[IssueEstimate]
    repo_id: Optional[int]
    is_epic: bool
    plus_ones: List[PlusOnes]
    pipeline: Optional[Pipeline]
    pipelines: Optional[List[Pipeline]]


class Issue(BaseModel):
    repo_id: int
    issue_number: int


class AddRemoveIssuesEpic(BaseModel):
    added_issues: List[Issue]
    removed_issues: List[Issue]


class AddRemoveIssue(BaseModel):
    added: List[Issue]
    removed: List[Issue]


class ReleaseReport(BaseModel):
    release_id: Base64String
    title: str
    description: str
    start_date: ISO8601DateString
    desired_end_date: ISO8601DateString
    created_at: ISO8601DateString
    closed_at: Optional[ISO8601DateString]
    state: ReportState
    repositories: Optional[List[int]]


class Dependency(BaseModel):
    blocking: Issue
    blocked: Issue


class Dependencies(BaseModel):
    dependencies: List[Dependency]


class MilestoneDate(BaseModel):
    start_date: ISO8601DateString


class Workspace(BaseModel):
    name: str
    description: Optional[str]
    id: Base64String
    repositories: List[int]


class Estimate(BaseModel):
    estimate: int


class PipelineIssue(BaseModel):
    issue_number: int
    estimate: Optional[IssueEstimate]
    position: Union[IssuePosition, int]
    is_epic: bool


class BoardPipeline(BaseModel):
    id: Base64String
    name: str
    issues: List[PipelineIssue]


class Board(BaseModel):
    pipelines: List[BoardPipeline]


class Epic(BaseModel):
    issue_number: int
    repo_id: int
    issue_url: URLString


class Epics(BaseModel):
    epic_issues: List[Epic]


class EpicData(BaseModel):
    total_epic_estimates: IssueEstimate
    estimate: Optional[IssueEstimate]
    pipeline: Pipeline
    pipelines: List[Pipeline]
    issues: List[Issue]


class RateLimit(BaseModel):
    limit: int
    used: int
    reset: Seconds



"""ZenHub release report issues methods."""

from typing import Iterable, List, Union

from ..models import AddRemoveIssue, Issue
from ..types import Base64String
from .base import BaseMixin


class ReleaseReportIssuesMixin(BaseMixin):
    def get_release_report_issues(
        self, release_id: Base64String
    ) -> Union[List[Issue], List[dict]]:
        """
        Get all the Issues for a Release Report.

        Parameters
        ----------
        release_id : Base64String
            The unique string identifier of the Release Report.

        Returns
        -------
        List of Issue or List of dictionaries
            See example response below.

        .. code-block:: python
            [
                { "repo_id": 103707262, "issue_number": 2 },
                { "repo_id": 103707262, "issue_number": 3 },
            ]

        Note
        ----
        https://github.com/ZenHubIO/API#get-all-the-issues-for-a-release-report
        """
        # GET /p1/reports/release/:release_id/issues
        url = f"/p1/reports/release/{release_id}/issues"
        if self._output_models:
            return [Issue.parse_obj(item) for item in self._get(url)]
        else:
            return [
                Issue.parse_obj(item).dict(include=item.keys())
                for item in self._get(url)
            ]

    def add_or_remove_issues_from_release_report(
        self,
        release_id: Base64String,
        add_issues: Iterable[Issue] = (),
        remove_issues: Iterable[Issue] = (),
    ) -> Union[AddRemoveIssue, dict]:
        """
        Add or Remove Issues to or from a Release Report.

        Adding and removing issues can be done in the same request by
        providing both ``add_issues`` and ``remove_issues`` parameters.

        Parameters
        ----------
        release_id : Base64String
            The unique string identifier of the Release Report.
        add_issues : Iterable of Issue
            An iterable of dictionaries with ``repo_id`` and ``issue_number``
            keys to add to the Release Report.
        remove_issues : Iterable of Issue
            An iterable of dictionaries with ``repo_id`` and ``issue_number``
            keys to remove from the Release Report.

        Returns
        -------
        AddRemoveIssue or dict
            The added or removed issues. See example response below.

        .. code-block:: python
            {
                "added": [{ "repo_id": 103707262, "issue_number": 3 }],
                "removed": [],
            }

        Note
        ----
        https://github.com/ZenHubIO/API#add-or-remove-issues-to-or-from-a-release-report
        """
        # PATCH /p1/reports/release/:release_id/issues
        url = f"/p1/reports/release/{release_id}/issues"
        body = {
            'add_issues': list(add_issues),
            'remove_issues': list(remove_issues),
        }
        data = self._patch(url, body=body)
        model = AddRemoveIssue.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )



"""Custom types."""
from typing_extensions import Literal

URLString = str
Base64String = str
ReportState = Literal["open", "closed"]
IssuePosition = Literal["top", "bottom"]
ISO8601DateString = str
Seconds = int
IssueEventType = Literal['estimateIssue', 'transferIssue']



"""ZenHub API."""
from typing import Optional

import requests

from ..types import URLString
from .dependencies import DependenciesMixin
from .epics import EpicsMixin
from .issues import IssuesMixin
from .milestones import MilestonesMixin
from .rate import RateMixin
from .release_report_issues import ReleaseReportIssuesMixin
from .release_reports import ReleaseReportsMixin
from .workspaces import WorkspacesMixin

# Constants
DEFAULT_BASE_URL: URLString = "https://api.zenhub.com"


class Zenhub(
    IssuesMixin,
    EpicsMixin,
    WorkspacesMixin,
    MilestonesMixin,
    DependenciesMixin,
    ReleaseReportsMixin,
    ReleaseReportIssuesMixin,
    RateMixin,
):
    """Zenhub API wrapper."""

    _HEADERS = {
        "Content-Type": "application/json",
        "User-Agent": "ZenHub Python Client",
    }

    def __init__(
        self,
        token: str,
        base_url: URLString = DEFAULT_BASE_URL,
        enterprise: int = 2,
        return_models: bool = False,
    ):
        """ZenHub API wrapper."""
        self._session = requests.Session()
        self._repo_id: Optional[int] = None

        if base_url == DEFAULT_BASE_URL:
            self._repo_id = 262640661  # Use ZenHub's default repo ID

        if enterprise == 3 and base_url != DEFAULT_BASE_URL:
            if base_url.endswith("/"):
                base_url = base_url + "api"
            else:
                base_url = base_url + "/api"

        self._base_url = base_url
        self._output_models = return_models

        # Setup
        self._session.headers.update(self._HEADERS)
        self._session.headers.update({"X-Authentication-Token": token})



"""ZenHub milestone methods."""
import datetime
from typing import Union

from ..models import MilestoneDate
from ..utils import date_to_string
from .base import BaseMixin


class MilestonesMixin(BaseMixin):
    def set_milestone_start_date(
        self,
        repo_id: int,
        milestone_number: int,
        start_date: datetime.datetime,
    ) -> Union[MilestoneDate, dict]:
        """
        Set milestone start date.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        milestone_number : int
            ID of the milestone, not its full name.
        start_date : datetime.datetime
            Start date of the milestone.

        Returns
        -------
        MilestoneDate or dict
            The milestone with the new start date. See example below.

        .. code-block:: python
            {
                "start_date": "2010-11-13T01:38:56.842Z",
            }

        Note
        ----
        https://github.com/ZenHubIO/API#set-milestone-start-date
        """
        self._repo_id = repo_id
        # POST /p1/repositories/:repo_id/milestones/:milestone_number/start_date
        url = (
            f"/p1/repositories/{repo_id}/milestones/"
            f"{milestone_number}/start_date"
        )
        body = {"start_date": date_to_string(start_date)}
        data = self._post(url, body)
        model = MilestoneDate.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def get_milestone_start_date(
        self, repo_id: int, milestone_number: int
    ) -> Union[MilestoneDate, dict]:
        """
        Get milestone start date.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        milestone_number : int
            ID of the milestone, not its full name.
        start_date : datetime.datetime
            Start date of the milestone.

        Returns
        -------
        MilestoneDate or dict
            The milestone with the current start date. See example below.

        .. code-block:: python
            {
                "start_date": "2010-11-13T01:38:56.842Z",
            }

        Note
        ----
        https://github.com/ZenHubIO/API#get-milestone-start-date
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/milestones/:milestone_number/start_date
        url = (
            f"/p1/repositories/{repo_id}/milestones/"
            f"{milestone_number}/start_date"
        )
        data = self._get(url)
        model = MilestoneDate.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )



# -----------------------------------------------------------------------------
# Copyright (c) 2022 Gonzalo Peña-Castellanos (@goanpeca)
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""ZenHub API."""

from typing import List, Union

from .core import Zenhub
from .exceptions import (
    APILimitError,
    InvalidTokenError,
    NotFoundError,
    ZenhubError,
)

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"


def _to_version_info(version: str) -> tuple:
    """Convert a version string to a number and string tuple."""
    parts: List[Union[str, int]] = []
    for part in version.split("."):
        if part.isnumeric():
            part = int(part)  # type: ignore

        parts.append(part)

    return tuple(parts)


VERSION_INFO = _to_version_info(__version__)

__all__ = [
    "APILimitError",
    "InvalidTokenError",
    "NotFoundError",
    "Zenhub",
    "ZenhubError",
]



"""ZenHub issues methods."""
from typing import List, Optional, Union

from ..models import (
    Estimate,
    EstimateIssueEvent,
    Event,
    IssueData,
    TransferIssueEvent,
)
from ..types import Base64String, IssuePosition
from .base import BaseMixin


class IssuesMixin(BaseMixin):
    def get_issue_data(
        self, repo_id: int, issue_number: int
    ) -> Union[IssueData, dict]:
        """
        Get the data for a specific issue.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            Repository issue number.

        Returns
        -------
        IssueData or dict
            The issue data dictionary. See example response below.

        .. code-block:: python
            {
                "estimate": {
                    "value": 8
                },
                "plus_ones": [
                    {
                        "created_at": "2015-12-11T18:43:22.296Z"
                    }
                ],
                "pipeline": {
                    "name": "QA",
                    "pipeline_id": "5d0a7a9741fd098f6b7f58a7",
                    "workspace_id": "5d0a7a9741fd098f6b7f58ac"
                },
                "pipelines": [
                    {
                        "name": "QA",
                        "pipeline_id": "5d0a7a9741fd098f6b7f58a7",
                        "workspace_id": "5d0a7a9741fd098f6b7f58ac"
                    },
                    {
                        "name": "Done",
                        "pipeline_id": "5d0a7cea41fd098f6b7f58b7",
                        "workspace_id": "5d0a7cea41fd098f6b7f58b8"
                    }
                ],
                "is_epic": True
            }

        Note
        ----
        - ``pipeline`` references the oldest Workspace pipeline this issue
          is in.
        - If an issue's status is closed, the pipeline value will describe the
          Pipeline that the issue was in prior to the issue being closed.
          The ZenHub API does not consider the "Closed" Pipeline to be a
          distinct Pipeline at this time and you should not use the Pipeline
          value to determine whether or not an issue is closed or open (use
          status instead).
        - Reopened issues might take up to one minute to show up in the
          correct Pipeline.
        - ``pipelines`` contains all pipelines in all Workspaces this issue
          is in.

        https://github.com/ZenHubIO/API#get-issue-data
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/issues/:issue_number
        url = f"/p1/repositories/{repo_id}/issues/{issue_number}"
        data = self._get(url)
        model = IssueData.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def get_issue_events(
        self, repo_id: int, issue_number: int
    ) -> Union[List[Event], List[dict]]:
        """
        Get the events for an issue.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            Repository issue number.

        Returns
        -------
        List of Event or List of dict
            See example response below.

        .. code-block:: python
            [
                {
                    "user_id": 16717,
                    "type": "estimateIssue",
                    "created_at": "2015-12-11T19:43:22.296Z",
                    "from_estimate": {
                        "value": 8
                    }
                },
                {
                    "user_id": 16717,
                    "type": "estimateIssue",
                    "created_at": "2015-12-11T18:43:22.296Z",
                    "from_estimate": {
                        "value": 4
                    },
                    "to_estimate": {
                        "value": 8
                    }
                },
                {
                    "user_id": 16717,
                    "type": "estimateIssue",
                    "created_at": "2015-12-11T13:43:22.296Z",
                    "to_estimate": {
                        "value": 4
                    }
                },
                {
                    "user_id": 16717,
                    "type": "transferIssue",
                    "created_at": "2015-12-11T12:43:22.296Z",
                    "from_pipeline": {
                        "name": "Backlog"
                    },
                    "to_pipeline": {
                        "name": "In progress"
                    },
                    "workspace_id": "5d0a7a9741fd098f6b7f58ac"
                },
                {
                    "user_id": 16717,
                    "type": "transferIssue",
                    "created_at": "2015-12-11T11:43:22.296Z",
                    "to_pipeline": {
                        "name": "Backlog"
                    }
                }
            ]

        Note
        ----
        - Returns issue events, sorted by creation time, most recent first.
        - Each event contains the User ID of the user who performed the
          change, the Creation Date of the event, and the event Type.
        - Type can be either estimateIssue or transferIssue. The values before
          and after the event are included in the event data.
        - transferIssue events include a workspace_id indicating in which
          Workspace the transfer occurred.

        https://github.com/ZenHubIO/API#get-issue-events
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/issues/:issue_number/events
        url = f"/p1/repositories/{repo_id}/issues/{issue_number}/events"
        events: List[dict] = self._get(url)  # type: ignore
        event_models: List[Event] = []
        for event in events:
            event_model: Optional[Event] = None
            if event["type"] == "estimateIssue":
                event_model = EstimateIssueEvent.parse_obj(event)
            elif event["type"] == "transferIssue":
                event_model = TransferIssueEvent.parse_obj(event)

            if event_model:
                event_models.append(event_model)

        return event_models if self._output_models else events

    def move_issue(
        self,
        workspace_id: Base64String,
        repo_id: int,
        issue_number: int,
        pipeline_id: Base64String,
        position: Union[int, IssuePosition],
    ) -> bool:
        """
        Moves an issue between Pipelines in a Workspace.

        Parameters
        ----------
        workspace_id : Base64String
            ID of the workspace.
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            Repository issue number.
        pipeline_id : Base64String
            ID of the pipeline, not its full name.
        position : int
            Can be specified as top or bottom, or a 0-based position in the
            Pipeline such as 1, which would be the second position in the
            Pipeline.

        Returns
        -------
        bool
            ``True`` if successful.

        Note
        ----
        https://github.com/ZenHubIO/API#move-an-issue-between-pipelines
        """
        self._repo_id = repo_id
        # POST /p2/workspaces/:workspace_id/repositories/:repo_id/issues/:issue_number/moves
        url = (
            f"/p2/workspaces/{workspace_id}/repositories/"
            f"{repo_id}/issues/{issue_number}/moves"
        )
        body = {"pipeline_id": pipeline_id, "position": position}
        return True if self._post(url, body) == {} else False

    def move_issue_in_oldest_workspace(
        self,
        repo_id: int,
        issue_number: int,
        pipeline_id: Base64String,
        position: Union[int, IssuePosition],
    ) -> bool:
        """
        Moves an issue between Pipelines in a Workspace.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            Repository issue number.
        pipeline_id : Base64String
            ID of the pipeline, not its full name.
        position : int or IssuePosition
            Can be specified as top or bottom, or a 0-based position in the
            Pipeline such as 1, which would be the second position in the
            Pipeline.

        Returns
        -------
        bool
            ``True`` if successful.

        Note
        ----
        https://github.com/ZenHubIO/API#move-an-issue-between-pipelines-in-the-oldest-workspace
        """
        self._repo_id = repo_id
        # POST /p1/repositories/:repo_id/issues/:issue_number/moves
        url = f"/p1/repositories/{repo_id}/issues/{issue_number}/moves"
        body = {"pipeline_id": pipeline_id, "position": position}
        return True if self._post(url, body) == {} else False

    def set_issue_estimate(
        self, repo_id: int, issue_number: int, estimate: int
    ) -> Union[Estimate, dict]:
        """
        Set Issue Estimate.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            Repository issue number.
        estimate : int
            The estimate for the issue.

        Returns
        -------
        Estimate or dict.
            See example response below.

        .. code-block:: python
            {
                "estimate": 15,
            }

        Note
        ----
        https://github.com/ZenHubIO/API#set-issue-estimate
        """
        self._repo_id = repo_id
        # PUT /p1/repositories/:repo_id/issues/:issue_number/estimate
        url = f"/p1/repositories/{repo_id}/issues/{issue_number}/estimate"
        body = {"estimate": estimate}
        data = self._put(url, body)
        model = Estimate.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )



"""ZenHub dependencies methods."""
from typing import Union

from ..models import Dependencies, Dependency
from .base import BaseMixin


class DependenciesMixin(BaseMixin):
    def get_dependencies(self, repo_id: int) -> Union[Dependencies, dict]:
        """
        Get Dependencies for a Repository.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        Dependencies or dict
            Dictionary of dependencies. See example below.

        .. code-block:: python
            {
                "dependencies": [
                    {
                        "blocking": {
                            "issue_number": 3953,
                            "repo_id": 1234567
                        },
                        "blocked": {
                            "issue_number": 1342,
                            "repo_id": 1234567
                        }
                    },
                    {
                        "blocking": {
                            "issue_number": 5,
                            "repo_id": 987
                        },
                        "blocked": {
                            "issue_number": 1342,
                            "repo_id": 1234567
                        }
                    }
                ]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#get-dependencies-for-a-repository
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/dependencies
        url = f"/p1/repositories/{repo_id}/dependencies"
        data = self._get(url)
        model = Dependencies.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def create_dependency(
        self,
        blocking_repo_id: int,
        blocking_issue_number: int,
        blocked_repo_id: int,
        blocked_issue_number: int,
    ) -> Union[Dependency, dict]:
        """
        Create a dependency.

        Parameters
        ----------
        blocking_repo_id: int
            ID of the repository, not its full name of the blocking issue.
        blocking_issue_number: int
            Blocking issue number.
        blocked_repo_id: int
            ID of the repository, not its full name of the blocked issue.
        blocked_issue_number: int
            Blocked issue number.

        Returns
        -------
        Dependency or dict
            Example response.

        .. code-block:: python
            {
                "blocking": {
                    "repo_id": 92563409,
                    "issue_number": 14
                },
                "blocked": {
                    "repo_id": 92563409,
                    "issue_number": 13
                }
            }

        Note
        ----
        https://github.com/ZenHubIO/API#create-a-dependency
        """
        self._repo_id = blocked_repo_id
        # POST /p1/dependencies
        url = "/p1/dependencies"
        body = {
            "blocking": {
                "repo_id": blocking_repo_id,
                "issue_number": blocking_issue_number,
            },
            "blocked": {
                "repo_id": blocked_repo_id,
                "issue_number": blocked_issue_number,
            },
        }
        data = self._post(url, body)
        model = Dependency.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def remove_dependency(
        self,
        blocking_repo_id: int,
        blocking_issue_number: int,
        blocked_repo_id: int,
        blocked_issue_number: int,
    ) -> bool:
        """
        Remove a dependency.

        Parameters
        ----------
        blocking_repo_id: int
            ID of the repository, not its full name of the blocking issue.
        blocking_issue_number: int
            Blocking issue number.
        blocked_repo_id: int
            ID of the repository, not its full name of the blocked issue.
        blocked_issue_number: int
            Blocked issue number.

        Returns
        -------
        bool
            ``True`` if the dependency was removed.

        Note
        ----
        https://github.com/ZenHubIO/API#remove-a-dependency
        """
        self._repo_id = blocked_repo_id
        # DELETE /p1/dependencies
        url = "/p1/dependencies"
        body = {
            "blocking": {
                "repo_id": blocking_repo_id,
                "issue_number": blocking_issue_number,
            },
            "blocked": {
                "repo_id": blocked_repo_id,
                "issue_number": blocked_issue_number,
            },
        }
        return True if self._delete(url, body) == {} else False



"""ZenHub dependencies methods."""
import datetime
from typing import Dict, Optional, Union

from requests.structures import CaseInsensitiveDict

from ..exceptions import ZenhubError
from ..models import RateLimit
from .base import BaseMixin


class RateMixin(BaseMixin):
    def _headers(self, repo_id: Optional[int] = None) -> CaseInsensitiveDict:
        """Return headers from example request.

        Parameters
        ----------
        repo_id : int, optional
            ID of the repository, not its full name. Default is `None`.

        Returns
        -------
        CaseInsensitiveDict
            Headers from request.
        """
        repo_id = repo_id or self._repo_id
        if repo_id is None:
            raise ZenhubError(
                "Need to make at least one request using the `repo_id` parameter!"
            )
        url = f'/p2/repositories/{repo_id}/workspaces'
        response = self._session.head(url=self._make_url(url))
        return response.headers

    def rate_limit(
        self, repo_id: Optional[int] = None
    ) -> Union[RateLimit, Dict[str, int]]:
        """API rate limit.

        A maximum of 100 requests per minute to the API.

        Parameters
        ----------
        repo_id : int, optional
            ID of the repository, not its full name. Default is `None`.

        Returns
        -------
        dict or RateLimit
            Dictionary with rate limit, available calls and time to reset
            limits in seconds.

        .. code-block:: python
            {'used': 3, 'limit': 100, 'reset': 58}

        Note
        ----
        https://github.com/ZenHubIO/API#api-rate-limit
        """
        headers = self._headers(repo_id)
        limit_used_string = headers.get('X-RateLimit-Used', -1)
        limit_allowed_string = headers.get('X-RateLimit-Limit', -1)
        limit_reset_time_string = headers.get('X-RateLimit-Reset', -1)
        server_date_string = headers.get('Date', -1)
        try:
            limit_used = int(limit_used_string)
        except Exception:
            limit_used = -1

        try:
            limit_allowed = int(limit_allowed_string)
        except Exception:
            limit_allowed = -1

        try:
            limit_reset_time = int(limit_reset_time_string)
        except Exception:
            limit_reset_time = -1

        if limit_reset_time != -1 or server_date_string != -1:
            date_reset = datetime.datetime.fromtimestamp(
                limit_reset_time, tz=datetime.timezone.utc
            )
            date_format = (
                '%a, %d %B %Y %H:%M:%S %Z'  # Mon, 30 May 2022 22:43:37 GMT
            )
            date_machine = datetime.datetime.strptime(
                server_date_string, date_format
            ).replace(tzinfo=datetime.timezone.utc)
            limit_reset = (date_reset - date_machine).seconds
        else:
            limit_reset = -1

        model = RateLimit(
            limit=limit_allowed, used=limit_used, reset=limit_reset
        )
        return model if self._output_models else model.dict()



"""ZenHub epics methods."""
from typing import Iterable, Union

from ..models import AddRemoveIssuesEpic, EpicData, Epics, Issue
from .base import BaseMixin


class EpicsMixin(BaseMixin):
    def get_epics(self, repo_id: int) -> Union[Epics, dict]:
        """
        Get all Epics for a repository.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.

        Returns
        -------
        Epics or dict
            See example response below.

        .. code-block:: python
            {
                "epic_issues": [
                    {
                        "issue_number": 3953,
                        "repo_id": 1234567,
                        "issue_url": "https://github.com/RepoOwner/RepoName/issues/3953"
                    },
                    {
                        "issue_number": 1342,
                        "repo_id": 1234567,
                        "issue_url": "https://github.com/RepoOwner/RepoName/issues/1342"
                    }
                ]
            }

        Note
        ----
        - The endpoint returns an array of the repository’s Epics. The issue
          number, repository ID, and GitHub issue URL is provided for each
          Epic.
        - If an issue is only an issue belonging to an Epic (and not a parent
          Epic), it is not considered an Epic and won’t be included in the return
          array.
        https://github.com/ZenHubIO/API#get-epics-for-a-repository
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/epics
        url = f"/p1/repositories/{repo_id}/epics"
        data = self._get(url)
        model = Epics.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def get_epic_data(
        self, repo_id: int, epic_id: int
    ) -> Union[EpicData, dict]:
        """
        Get all Epics for a repository.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        epic_id : int
            Github issue number of the epic.

        Returns
        -------
        EpicData or dict
            See example response below.

        .. code-block:: python
            {
                "total_epic_estimates": {"value": 60},
                "estimate": {"value": 10},
                "pipeline": {
                    "workspace_id": "5d0a7a9741fd098f6b7f58ac",
                    "name": "Backlog",
                    "pipeline_id": "5d0a7a9741fd098f6b7f58a8"
                },
                "pipelines": [
                    {
                        "workspace_id": "5d0a7a9741fd098f6b7f58ac",
                        "name": "Backlog",
                        "pipeline_id": "5d0a7a9741fd098f6b7f58a8"
                    },
                    {
                        "workspace_id": "5d0a7cea41fd098f6b7f58b8",
                        "name": "In Progress",
                        "pipeline_id": "5d0a7cea41fd098f6b7f58b5"
                    }
                ],
                "issues": [
                    {
                        "issue_number": 3161,
                        "is_epic": True,
                        "repo_id": 1099029,
                        "estimate": {"value": 40},
                        "pipelines": [
                            {
                                "workspace_id": "5d0a7a9741fd098f6b7f58ac",
                                "name": "Backlog",
                                "pipeline_id": "5d0a7a9741fd098f6b7f58a8"
                            },
                            {
                                "workspace_id": "5d0a7cea41fd098f6b7f58b8",
                                "name": "In Progress",
                                "pipeline_id": "5d0a7cea41fd098f6b7f58b5"
                            }
                        ],
                        "pipeline": {
                            "workspace_id": "5d0a7a9741fd098f6b7f58ac",
                            "name": "Backlog",
                            "pipeline_id": "5d0a7a9741fd098f6b7f58a8"
                        }
                    },
                    {
                        "issue_number": 2,
                        "is_epic": False,
                        "repo_id": 1234567,
                        "estimate": {"value": 10},
                        "pipelines": [
                            {
                                "workspace_id": "5d0a7a9741fd098f6b7f58ac",
                                "name": "Backlog",
                                "pipeline_id": "5d0a7a9741fd098f6b7f58a8"
                            },
                            {
                                "workspace_id": "5d0a7cea41fd098f6b7f58b8",
                                "name": "In Progress",
                                "pipeline_id": "5d0a7cea41fd098f6b7f58b5"
                            }
                        ],
                        "pipeline": {
                            "workspace_id": "5d0a7a9741fd098f6b7f58ac",
                            "name": "Backlog",
                            "pipeline_id": "5d0a7a9741fd098f6b7f58a8"
                        }
                    }
                ]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#get-epic-data
        """
        self._repo_id = repo_id
        # GET /p1/repositories/:repo_id/epics/:epic_id
        url = f"/p1/repositories/{repo_id}/epics/{epic_id}"
        data = self._get(url)
        model = EpicData.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )

    def convert_epic_to_issue(self, repo_id: int, issue_number: int) -> bool:
        """
        Converts an Epic back to a regular issue.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            The number of the issue to be converted

        Returns
        -------
        bool
            ``True`` if successful.

        Note
        ----
        https://github.com/ZenHubIO/API#convert-an-epic-to-an-issue
        """
        self._repo_id = repo_id
        # POST /p1/repositories/:repo_id/epics/:issue_number/convert_to_issue
        url = (
            f"/p1/repositories/{repo_id}/epics/{issue_number}/convert_to_issue"
        )
        data = self._post(url)
        return True if data == {} else False

    def convert_issue_to_epic(
        self, repo_id: int, issue_number: int, issues: Iterable[Issue] = ()
    ) -> bool:
        """
        Converts an issue to an Epic, along with any issues that should be
        part of it.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            The number of the issue to be converted
        issues: Iterable[Issue]
            Add issues to epic.

        Returns
        -------
        bool
            ``True`` if successful.

        Note
        ----
        https://github.com/ZenHubIO/API#convert-issue-to-epic
        """
        self._repo_id = repo_id
        # POST /p1/repositories/:repo_id/issues/:issue_number/convert_to_epic
        url = (
            f"/p1/repositories/{repo_id}/issues/{issue_number}/convert_to_epic"
        )
        body = {"issues": issues}
        data = self._post(url, body=body)
        return bool(data)

    def add_or_remove_issues_to_epic(
        self,
        repo_id: int,
        issue_number: int,
        remove_issues: Iterable[Issue] = (),
        add_issues: Iterable[Issue] = (),
    ) -> Union[AddRemoveIssuesEpic, dict]:
        """
        Bulk add or remove issues to an Epic.

        The result returns which issue was added or removed from the Epic.

        Parameters
        ----------
        repo_id : int
            ID of the repository, not its full name.
        issue_number : int
            The number of the epic.
        remove_issues: Iterable[Issue]
            Issues to remove from epic.
        add_issues: Iterable[Issue]
            Issues to add to epic.

        Returns
        -------
        AddRemoveIssuesEpic or dict
            Example response.

        .. code-block:: python
            {
                "removed_issues": [
                    {"repo_id": 3887883, "issue_number": 3}
                ],
                "added_issues": [
                    {"repo_id": 3887883, "issue_number": 2},
                    {"repo_id": 3887883, "issue_number": 1}
                ]
            }

        Note
        ----
        https://github.com/ZenHubIO/API#add-or-remove-issues-to-epic
        """
        self._repo_id = repo_id
        # POST /p1/repositories/:repo_id/epics/:issue_number/update_issues
        url = f"/p1/repositories/{repo_id}/epics/{issue_number}/update_issues"
        body = {
            "remove_issues": list(remove_issues),
            "add_issues": list(add_issues),
        }
        data = self._post(url, body=body)
        model = AddRemoveIssuesEpic.parse_obj(data)
        return (
            model if self._output_models else model.dict(include=data.keys())
        )



"""ZenHub API Custom Exeptions."""


class ZenhubError(Exception):
    pass


class InvalidTokenError(ZenhubError):
    pass


class APILimitError(ZenhubError):
    pass


class NotFoundError(ZenhubError):
    pass
